 // Name: crush.c
 // 24.09.96

 #include <alloc.h>
 #include <assert.h>
 #include <fcntl.h>
 #include <io.h>
 #include <string.h>
 #include <values.h>
 #include <conio.h>
 #include <stdio.h>
 #include <ctype.h>

 #include "crush.h"

 void help(void);

 //--------------------------------------------------------------------------
 int l_scal_code(scalType op_type)
   {
     return (op_type == float_type) << 1;
   }

 //--------------------------------------------------------------------------
 int r_scal_code(scalType op_type)
   {
     return (op_type == float_type);
   }

 //--------------------------------------------------------------------------
 int l_scal_type(scalCode code)
   {
     return (code & 2) ? float_type: int_type;
   }

 //--------------------------------------------------------------------------
 int r_scal_type(scalCode code)
   {
     return (code & 1) ? float_type: int_type;
   }

 //--------------------------------------------------------------------------
 long load_table(char *name, char **table)
   {
     int handle, i;
     long len;

     if ((handle = _open(name, O_RDONLY)) < 0)
       {
         printf("訡  䠩 '%s' !\n", name);
         abort();
       }

     len = filelength(handle);
     if (!len)
       {
         _close(handle);
         return 0;
       }

     assert(*table = malloc(len));

     i = len / (MAXINT-1);
     do
       if (_read(handle, *table, MAXINT-1) == -1)
         {
           printf("訡 ⥭ 䠩 %s!\n", name);
           free(*table);
           return 0;
         }
     while (i--);

     _close(handle);
     return len;
   } // load_table


 //--------------------------------------------------------------------------
 char *name;
 char *exts[] = { ".ins", ".nam", ".opd", ".cnt", ".dmp" };

 Instr   *instr_tab;
 int     instr_num;

 Ident   *ident_tab;
 int     ident_num;

 Operand *oper_tab;
 int     oper_num;

 Const   *const_tab;
 int     const_num;

 char *mem;
 int mem_size = 0;

 char stack[stack_size];
 int stack_top = 0;

 int debug = 0;
 int dump  = 0;

 //--------------------------------------------------------------------------
 int main(int arg_num, char *args[])
   {
     int tab_num, i;
     if (arg_num<2)
       {
         puts("Example: crush.exe <input_file>");
         return 0;
       }

     if (args[2][0] == 'd') debug = 1;

     // --------------- ࠧ饭   ⠡ --------------------------
     // -------- 権, ६, ࠭  ⠭ ---------------

     assert(name = malloc(strlen(args[1])+4));
     for (tab_num = 0; tab_num < 4; tab_num++)
       {
         strcpy(name, args[1]);
         strcat(name, exts[tab_num]);
         switch (tab_num)
           {
             case 0:
               instr_num = load_table(name, &instr_tab) / sizeof(Instr);
               break;

             case 1:
               ident_num = load_table(name, &ident_tab) / sizeof(Ident);
               break;

             case 2:
               oper_num = load_table(name, &oper_tab) / sizeof(Operand);
               break;

             case 3:
               const_num = load_table(name, &const_tab) / sizeof(Const);
               break;
           }

       } // for (tab_num = 0; tab_num < 4; tab_num++)

     //---------- ࠧ饭   ᪨ ६ ---------------
     for (i = 0; i < ident_num; i++)
       mem_size += ident_tab[i].len ? ident_tab[i].len : 1;
     if (mem_size)
       assert(mem = malloc(mem_size <<= 1));

     // -------------------  樨 --------------------------
     i = 0;
     while (debug)
       {
         char com[41], *cur;
         int  start_addr, last_addr, addr, num;

         printf("\n- current ip = %d\npult> ", i);
         scanf("%40s", com);
         cur = com;

         switch (tolower(*cur++))
           {
             case 'd':
               if (str2uns(&cur, &i) && !*cur)
                 if (i < instr_num)
                   {
                     printf("- start program with memory dump at instruction %d:\n", i);
                     debug = 0;
                     dump  = 1;
                   }
                 else puts("- instruction number too big !");
               else puts("- syntax error in command !");
               break;

             case 'g':
               if (str2uns(&cur, &i) && !*cur)
                 if (i < instr_num)
                   {
                     printf("- start program at instruction %d:\n", i);
                     debug = 0;
                   }
                 else puts("- instruction number too big !");
               else puts("- syntax error in command !");
               break;

             case 'h':
               if (*cur) puts("- syntax error in command !");
               help();
               break;

             case 'm':
               start_addr = 0;
               if (str2uns(&cur, &start_addr))
                 {
                   puts("- type memory contens:");
                   if (!*cur) last_addr = start_addr;
                   else if (*cur++ == '-')
                     {
                       last_addr = mem_size;
                       if (!str2uns(&cur, &last_addr) || *cur)
                         {
                           puts("- syntax error in command !");
                           break;
                         }
                     }
                   else
                     {
                       puts("- syntax error in command !");
                       break;
                     }

                   num = 0;
                   for (addr = start_addr; addr <= last_addr; addr += 2)
                     {
                       printf("addess %3d: int %d \tfloat %e\n", addr,
                            *(int *)&mem[addr], *(float *)&mem[addr]);

                       if (++num > 23)
                         {
                           num = 0;
                           printf("- continue ? ");
                           if (tolower(getche()) == 'n') break;
                         }

                     }

                 } // if (str2uns(&cur, &start_addr))
               else puts("- syntax error in command !");
               break;

             case 'p':
               if (!*cur)
                 {
                   puts("- print instruction:");
                   print_instr(i);
                 }
               else puts("- syntax error in command !");
               break;

             case 's':
               start_addr = 0;
               last_addr = mem_size;
               if (str2uns(&cur, &start_addr))
                 {
                   puts("- set value to memory:");
                   if (*cur == '=') last_addr = start_addr;
                   else if (*cur++ != '-' || !str2uns(&cur, &last_addr))
                     {
                       puts("- syntax error in command !");
                       break;
                     }

                   num = 0;
                   if (*cur++ != '=' || !str2uns(&cur, &num) || *cur)
                     {
                       puts("- syntax error in command !");
                       break;
                     }

                   for (addr = start_addr; addr <= last_addr; addr += 2)
                     {
                       *(int *)&mem[addr] = num;
                       printf("addess %3d: %d \n", addr, num);
                     }

                 } // if (str2uns(&cur, &start_addr))
               else puts("- syntax error in command !");
               break;

             case 't':
               if (str2uns(&cur, &i) && !*cur)
                 if (i < instr_num)
                   {
                     puts("- trace instruction:");
                     print_instr(i);
                     i = trace(i);
                   }
                 else puts("- instruction number too big !");
               else puts("- syntax error in command !");
               break;

             case 'q':
               if (!*cur)
                 {
                   i = instr_num;
                   debug = 0;
                 }
               else puts("- syntax error in command !");
               break;

             default:
               puts("- syntax error in command !");
               help();

           } // switch (tolower(*cur++))

       } // while (debug)

     while (i < instr_num) i = trace(i);

     if (dump)
       {
         int handle;

         printf("\n- memory dump save at name '%s'd0\n", name);
         strcpy(name, args[1]);
         strcat(name, exts[4]);

         if ((handle = _creat(name, 32)) < 0)
           printf("訡 ᮧ 䠩 '%s' !\n", name);
         else if (_write(handle, mem, mem_size) == -1)
             printf("訡   䠩 %s!\n", name);

         _close(handle);
       } // if (dump)

     free(name);
     free(instr_tab);
     free(ident_tab);
     free(oper_tab);
     free(const_tab);
     free(mem);

     return 0;
   }
