 #ifndef _CRUSH_H
 #define _CRUSH_H

 #include <stdio.h>
 #include <stdlib.h>

 typedef enum
   {
     addOpc,       assOpc,         divOpc,         eqOpc,
     emptyOpc,     exitOpc,        geOpc,          gotoOpc,
     gtOpc,        ifOpc,          inOpc,          indexOpc,
     labelOpc,     leOpc,          ltOpc,          minOpc,
     modOpc,       multOpc,        neOpc,          outOpc,
     skipOutOpc,   spaceOutOpc,    subOpc,         tabOutOpc
   }
 opcType;

 typedef enum
   {
     int_type,
     float_type
   }
 scalType;

 typedef enum
   {
     nameVarOpd, //  ६  ⠡ 
     tmpVarOpd,  // ஬筠 ६  ࠦ
     constOpd,   // ⠭ ࠭
     labelOpd,    // ࠭ - ⪠
     indexVarOpd  // ஢ ६
   }
 opdType;

 #include "transl.h"

 typedef struct Ident
   {
     char     name[name_len];  //  ६
     scalType type;            // ᪠ ⨯ ६
     int      len;             //   ( = 0, ᫨ ᪠)
     int      addr;            //   䨧᪮ 
   }
 Ident;

 typedef struct Operand
   {
     opdType type;        // ⨯ ࠭
     union
       {
         int temp_num;
         int ident_num;
         int const_num;
         int instr_num;
       }
     val;
   }
 Operand;

 typedef struct Const
   {
     scalType type;      // ⨯ ⠭
     union
       {
         int   inum;    // 楫 ⠭
         float fnum;    // ⢨⥫쭠 ⠭
       }
     val;
   }
 Const;

 typedef struct Temp
   {
     int          id;              //  ६
     scalType     type;            // ᪠ ⨯ ६
     int          len;             //   ( = 0, ᫨ ᪠)
     int          addr;            //   ⥪
     struct Temp *next;
     struct Temp *prev;
   }
 Temp;

 #define error(str) \
   printf("\nError in line %d: %s !\n", instr_tab[i].line, str), abort();

 #define stack_size 1024

 extern Instr   *instr_tab;
 extern int     instr_num;

 extern Ident   *ident_tab;
 extern int     ident_num;

 extern Operand *oper_tab;
 extern int     oper_num;

 extern Const   *const_tab;
 extern int     const_num;

 extern Temp    *temp_list, *last_temp;
 extern Temp    *index_list, *last_index;

 extern char *mem;
 extern int mem_size;

 extern char stack[stack_size];
 extern int stack_top;

 void  ins_temp(Temp *ptr);
 Temp *get_temp(int id);

 void  ins_index(Temp *ptr);
 Temp *get_index(int id);

 typedef enum
   {
     int_int,
     int_float,
     float_int,
     float_float
   }
 scalCode;

 int l_scal_code(scalType op_type);
 int r_scal_code(scalType op_type);

 int l_scal_type(scalCode op_type);
 int r_scal_type(scalCode op_type);

 void addition(int i);
 void assign(int i);
 void divide(int i);
 void eq(int i);
 void ge(int i);
 void gt(int i);
 void input(int i);
 void index(int i);
 void le(int i);
 void lt(int i);
 void neg(int i);
 void mod(int i);
 void multiple(int i);
 void ne(int i);
 void output(int i);
 void substr(int i);
 void symout(int i, char sym);

 int   str2int(char *str, int *rez_ptr);
 int   str2uns(char **str, int *rez);
 int   str2float(char *str, float *rez_ptr);
 int   trace(int i);
 void  help();
 void  print_instr(int  i);

 #endif